(* ::Package:: *)

(* ::Title:: *)
(*Mathematica package to parse LHA CT10.info and CT10_**** files*)


(* ::Text:: *)
(*Original Version: 3 Feb 2014*)
(*History : *)


(* ::Section:: *)
(*Setup Package*)


BeginPackage["pdfParseLHA`"];

Print["==============================================================="];
Print[" "];
Print[" - pdfParseLHA - "];
Print["Version:  1.0"];
Print["Authors: E.J. Godat, D.B. Clark & F.I. Olness"];
Print[" "];
Print["Please cite: **************"];
Print[" "];
Print[" "];
Print["For a list of avialable commands, enter: ?pdf*"];
Print[" "];
Print["==============================================================="];


(***********************************************************************)
pdfFunctionLHA::usage="pdfFunctionLHA[setNumber,flavor,x,Q]: This function returns the value of the pdf for the .dat file corresponding to \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\), for 
the given flavor and value of Bjorken x and Q.

\!\(\*
StyleBox[\"Warning\",\nFontSlant->\"Italic\"]\): The results of this function are only reliable between the maximmum and minimum values of x and Q in the .info file.";

(***********************************************************************)

pdfParseLHA::usage="pdfParseLHA[fileNameInfo, fileNameData]: This funciton reads an info file and a data file into memory and checks that the number and the order of the flavors are the same in both files";

(***********************************************************************)

pdfFamilyParseLHA::usage="pdfFamilyParseLHA[path,[fileType]]: This function reads all the files of type \!\(\*
StyleBox[\"fileType\",\nFontSlant->\"Italic\"]\) in the 
directory path and stores them in memory. The function returns a list of set numbers that can be used to define a list. 
The optional input fileType has a default value of \"*.dat\".

Example:
  pdfFamilyParseLHA[\"MyGrids\",\"ct10*dat\"] reads all .dat files in the subdirectory \"MyGrids\" beginning with \"ct10\" into memory.";

(***********************************************************************)

pdfSetListLHA::usage="pdfSetListLHA: This global variable prints the list of the loaded .dat grid files, their setNumber, maximal number of quark flavors, 
and number of valence flavors in each PDF set.";

(***********************************************************************)

pdfResetLHA::usage="pdfResetLHA[]: This funtion deletes all .dat files from memory and resets all the internal variables in the package.

\!\(\*
StyleBox[\"Note\",\nFontSlant->\"Italic\"]\): This function does not accept any inputs.";

(***********************************************************************)

pdfFlavorLHA::usage="pdfFlavorLHA[flavor]: This function returns a string with the name of the PDF flavor. 

Example:
pdfFlavorLHA[0], pdfFlavorLHA[1], pdfFlavorLHA[2] will return \"gluon\", \"down\", \"up\" for the gluon, down quark, and up quark PDFs.";

(***********************************************************************)

pdfXminLHA::usage="pdfXminLHA[setNumber]: This function returns the minimum x value in the PDF grid \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\).";

(***********************************************************************)

pdfGetXlistLHA::usage="pdfGetXlistLHA[setNumber]: This function returns the grid values in x from the PDF set \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\).";

(***********************************************************************)

pdfGetQlistLHA::usage="pdfGetQlistLHA[setNumber]: This function returns the grid values in Q from the PDF set \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\).";

(***********************************************************************)

pdfGetTableLHA::usage="pdfGetTableLHA[setNumber]: This function returns the values sorted by flavor, q ,then x from the PDF set \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\).  If looking for a specific value, use the pdfGetValueLHA function.";

(***********************************************************************)

pdfGetValueLHA::usage="pdfGetValueLHA[setNumber,flavor,q value, x value]: This function returns the grid value for a specific flavor, q indicie and x indicie from the PDF set \!\(\*
StyleBox[\"setNumber\",\nFontSlant->\"Italic\"]\).

It is possible to input the ranges in the form Integer;;Integer to produce a list of values.

\!\(\*
StyleBox[\"Note\",\nFontSlant->\"Italic\"]\): The indicies for q and x values must be Integers or a range of Integers.";

(***********************************************************************)

pdfGetInfoLHA::usage="pdfGetInfoLHA[]: This function returns the info read from the .info file

\!\(\*
StyleBox[\"Note\",\nFontSlant->\"Italic\"]\): This function does not accept any inputs.";

(***********************************************************************)

pdfGetAlphaValuesLHA::usage="pdfGetAlphaValuesLHA[]: This function returns the list of AlphaS values from the .info file

\!\(\*
StyleBox[\"Note\",\nFontSlant->\"Italic\"]\): This function does not accept any inputs.";

(***********************************************************************)

pdfAlphaSLHA::usage="pdfAlphaSLHA[Q]:This function returns the value of \!\(\*SubscriptBox[\(\[Alpha]\), \(S\)]\) at hard scattering energy \!\(\*
StyleBox[\"Q\",\nFontSlant->\"Italic\"]\) when this information is available in the .info file.";

(***********************************************************************)

Begin["`Private`"]; 


(* ::Section:: *)
(*Main Parsing Functions*)



(* INITIALIZATION *)

	Clear[flavorlist, flavorcheck, readlist, datalist, AlphaQlist, AlphaValslist, xlist, qlist, bigTable,pdfTableData,nSetCount,pdfSetListLHA];
	nSetCount=0;
	pdfSetListLHA={};
	alphalist={};
	pdfxmin[_]=False;(*for error control*)
	pdfTableData[_]=False;(*for error control*)

(* The above are global variables *)



(* ::Code:: *)
(********************************** MAIN PARSING ROUTINE**********************************)

pdfParseLHA[fileNameInfo_,fileNameData_]:=Module[{},

readInfo[fileNameInfo];
readData[fileNameData];

(*Checks that the flavors in the info file matches the flavors in the data files*)
If[flavorlist!=flavorcheck,Print["Error, Flavors from info file do not match flavors in data file"]; Return[]];

]

(********************************************************)

(*Reads LHA .info files*)
readInfo[fileName_]:=Module[{stream,mm,tmpList,tempStream,miscList},

If[!StringQ[fileName],Print["Error, File Name was not in the form of a String"];Return[]];
If[!FileExistsQ[fileName],Print["Error, file does not exist"];Return[]];
If[!StringMatchQ[fileName,"*.info"],Print["Error, not .info"];Return[]];

stream=OpenRead[fileName];
SetStreamPosition[stream,0];
readlist={"NumFlavors","XMin","XMax","QMin","QMax","MZ","AlphaS_MZ","AlphaS_OrderQCD"}; (*List of the single numeric values from the info file*)
datalist={}; 

(*Loop through the list of variables and save them in the datalist*)
For[i=1,i<=Length[readlist],i++,
SetStreamPosition[stream,0];
Find[stream,readlist[[i]],RecordSeparators->": "];
SetStreamPosition[stream,StreamPosition[stream]+1];(*the +1 advances from the space to the numeric value*)
AppendTo[datalist,N[Read[stream,Number]]]
];

(*Read in the flavors from the info file*)
SetStreamPosition[stream,0];
Find[stream,"Flavors",RecordSeparators->": "];
SetStreamPosition[stream,StreamPosition[stream]+3]; (*The +3 advances past the { at the start of the list in the info file*)
StreamPosition[stream];
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number,RecordSeparators->","];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]+1];
mm++,
{i,1,200}
];
flavorlist=Table[tmpList[i],{i,1,mm}];(*List of flavors*)

(*Read in the AlphaS_Qs list*)
SetStreamPosition[stream,0];
Find[stream,"AlphaS_Qs",RecordSeparators->": "];
SetStreamPosition[stream,StreamPosition[stream]+3];(*The +3 advances past the { at the start of the list in the info file*)
StreamPosition[stream];
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number,RecordSeparators->","];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]+1];
mm++,
{i,1,200}
];
AlphaQlist=Table[tmpList[i],{i,1,mm}]; (*AlphaQ list*)

SetStreamPosition[stream,0];
Find[stream,"AlphaS_Vals",RecordSeparators->": "];
SetStreamPosition[stream,StreamPosition[stream]+3];(*The +3 advances past the { at the start of the list in the info file*)
StreamPosition[stream];
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number,RecordSeparators->","];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]+1];
mm++,
{i,1,200}
];
AlphaValslist=Table[tmpList[i],{i,1,mm}]; (*Alpha Value list*)

alphalist={AlphaQlist,AlphaValslist};

Close[stream];
Print["Successfully read ",fileName,"."];
]


(********************************************************)

(*Reads LHA .dat files*)
readData[fileName_]:=Module[{stream,mm,tmpList,tempStream,bigTableRow},

If[!StringQ[fileName],Print["Error, File Name was not in the form of a String"];Return[]];
If[!FileExistsQ[fileName],Print["Error, file does not exist"];Return[]];
If[!StringMatchQ[fileName,"*.dat"],Print["Error, not .dat"];Return[]];

stream=OpenRead[fileName]; 
SetStreamPosition[stream,0];

(*Reads the list of x values after the --- indicator*)
Find[stream,"---"];
SetStreamPosition[stream,StreamPosition[stream]];
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]];
mm++,
{i,1,2000}
];
xlist=Table[tmpList[i],{i,1,mm}];(*List of x values*)

(*Reads the list of q values*)
StreamPosition[stream];
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]];
mm++,
{i,1,2000}
];
qlist=Table[tmpList[i],{i,1,mm}]; (*List of q values*)

(*Reads the list of flavors*)
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]];
mm++,
{i,1,2000}
];
flavorcheck=Table[tmpList[i],{i,1,mm}];(*List of flavors, this should match the flavors read from the info file*)

(*Reads the list of pdf values*)
bigTable={};
For[i=1,i<=Length[xlist]*Length[qlist],i++,(*The length of the pdf value table is determined by the product of the number of x and q values*)
tempStream=StringToStream[Read[stream,String]];
mm=0;
Do[
  tmpList[i]=Read[tempStream,Number];
If[tmpList[i]==EndOfFile,Break[]];
SetStreamPosition[tempStream,StreamPosition[tempStream]];
mm++,
{i,1,2000}
];
bigTableRow=Table[tmpList[i],{i,1,mm}];(*Each row should contain a value for each flavor of a given x and given q*)
AppendTo[bigTable,bigTableRow];
];

(*Partitions and reorganizes bigTable*)

numbX=Length[xlist];(*Number of X values in the xlist*)
numbQ=Length[qlist];(*Number of Q values in the qlist*)

bigTableCut=Partition[Partition[bigTable,{numbQ}],{numbX}][[1]];(*Table divided into the numbQ, numbX, flavor*)
(*This leaves a table with indicies refering to (x,q,flavor)*)
bigTableCutTrans=Transpose[bigTableCut,{3,2,1}];(*This reorganizes the indicies into (flavor,q,x)*)

list=bigTableCutTrans;
miscList={Length[flavorcheck],5,3};

nSetCount=nSetCount+1;
pdfxmin[nSetCount]=datalist[[2]];
pdfTableData[nSetCount]={xlist,qlist,list,miscList};

AppendTo[pdfSetListLHA,{nSetCount,fileName}];



Close[stream];
Print["Successfully read ",fileName,"."];
];

(********************************** END MAIN PARSING ROUTINE**********************************)






(*pdfResetLHA: no input, output is to reset memory and all global variables in the package*)
pdfResetLHA[]:=Module[{},
	Clear[flavorlist, flavorcheck, readlist, datalist, AlphaQlist, AlphaValslist, xlist, qlist, bigTable,pdfTableData,nSetCount,pdfSetListLHA];
	nSetCount=0;
	pdfSetListLHA={};
	alphalist={};
	pdfxmin[_]=False;(*for error control*)
	pdfTableData[_]=False;(*for error control*)
	Print["All internal variables have been reset."]
  ];






(*pdfFamilyParseLHA: input is a directory containing a single .info file and miltiple .dat files (optional input is filetype-->default is "*.dat")*)
(*                     all .dat files are stored in memory and output is a list containing the setNumbers of the .dat files that were stored *)
pdfFamilyParseLHA[path_?StringQ,type_:"*.dat"]:=
	Module[{currDir,datfileList,output,fLength,startSet,endSet,infoFile},
	currDir=Directory[];
	SetDirectory[path];
	infoFile=FileNames[path<>"/"<>"*.info"];
	If[Length[infoFile]!=1,Print["Error, Missing or Multiple Info Files"]; Return[],infoFile=infoFile[[1]]]
	readInfo[infoFile];
	datfileList=FileNames[path<>"/"<>type];
	fLength=Length[datfileList];
	startSet=nSetCount+1;
	Do[readData[datfileList[[i]]],{i,1,fLength}];
	endSet=nSetCount;
	Print["Included ",endSet-startSet+1," files in the PDF family."];
	SetDirectory[currDir];
	output=Range[startSet,endSet];
	Return[output];
  ];





(* ::Section:: *)
(*Interpolation*)


(* ::Subsection:: *)
(*This is a 4 - point (cubic) interpolation funtion.*)



(*interpol4: a four-point Lagrange interpolation used throughout this package-->input is a point and a list (data) of the 4 points in x *)
(*           and Q surrounding the point,output is the value interpolated at the input point *)
interpol4[xIn_,data_]:=Module[
	{x,xVec,qVec,x0,x1,x2,x3,q0,q1,q2,q3,c0,c1,c2,c3,output},
	{xVec,qVec}=Transpose[data];
	x=xIn;
	{x0,x1,x2,x3}=xVec;
	{q0,q1,q2,q3}=qVec;
	c0=((x-x1)/(x0-x1)) ((x-x2)/(x0-x2)) ((x-x3)/(x0-x3));
	c1=((x-x0)/(x1-x0)) ((x-x2)/(x1-x2)) ((x-x3)/(x1-x3));
	c2=((x-x0)/(x2-x0)) ((x-x1)/(x2-x1)) ((x-x3)/(x2-x3));
	c3=((x-x0)/(x3-x0)) ((x-x1)/(x3-x1)) ((x-x2)/(x3-x2));
	output=q0 c0+q1 c1+q2 c2+q3 c3;
	If[output<0,Return[0],Return[output]]
  ];





(* ::Subsection:: *)
(*Interpolation Subfunctions*)



(*bigger: compares tow points xin and xintp and gives True if xin is greater than or equal to xintp*)
bigger[xin_,xintp_]:=(xin>=xintp);
SetAttributes[bigger,Listable];


(*findXindex:locates the point in a grid xlist that is immediately before the point x*)
(*input is point x and grid xlist, output is point in grid immdiately before x*)
findXindex[x_,xlist_]:=Module[{pos,xmin,xmax,r,xmis},
	xmin=xlist[[2]];
	r=Length[xlist];
	xmis=xlist[[r-1]];
	If[x<=xmin,Return[4],If[x>=xmis,Return[(r-1)],Return[Position[bigger[xlist,x],True]//First//First]];];
  ];


(*findQindex:locates the point in a grid qlist that is immediately before the point q*)
(*input is point q and grid qlist, output is point in grid immdiately before q*)
(*this is a unique function from findXindex as the structure of the two lists is different!!!*)
findQindex[q_,qlist_]:=Module[{pos,qmin,qmax,r,qmis},
	qmin=qlist[[2]];
	r=Length[qlist];
	qmis=qlist[[r-1]];
	If[q<=qmin,Return[3],If[q>=qmis,Return[(r-1)],Return[Position[bigger[qlist,q],True]//First//First]];];
  ];


(*getXdata && getQdata: construct the data lists for the interpolation for the appropriate points*)
(*these functions are unique*)
getXdata[xi_,qi_,flav_,xlist_,grid_]:=Transpose[{xlist[[xi-2;;xi+1]],grid[[flav,qi,xi-2;;xi+1]]}];
getQdata[xi_,qi_,flav_,qlist_,grid_]:=Transpose[{qlist[[qi-2;;qi+1]],grid[[flav,qi-2;;qi+1,xi]]}];


(*doQinterp: does the full interpolation over the 4 Q points starting from xi and qi for the flavor and grids *)
doQinterp[x_,xi_,qi_,flav_,xlist_,qlist_,grid_]:=
	Module[{pVec,qVec,output},
	pVec={
	interpol4[x,getXdata[xi,qi-2,flav,xlist,grid]],
	interpol4[x,getXdata[xi,qi-1,flav,xlist,grid]],
	interpol4[x,getXdata[xi,qi,flav,xlist,grid]],
	interpol4[x,getXdata[xi,qi+1,flav,xlist,grid]]};
	qVec=qlist[[qi-2;;qi+1]];
	output={qVec,pVec}//Transpose;
	Return[output]
  ];


(*fullinterp:does the interpolation over all the points in 4 directions outputs the value at that point*)
fullinterp[x_?NumericQ,q_?NumericQ,flav_,list_]:=
	Module[{xi,qi,qData,output,xlist,qlist,grid},
	xlist=list[[1]];
	qlist=list[[2]];
	grid=list[[3]];
	xi=findXindex[x,xlist];
	qi=findQindex[q,qlist];
	qData=doQinterp[x,xi,qi,flav,xlist,qlist,grid];
	output=interpol4[q,qData];
	Return[output];
  ];





(* ::Section:: *)
(*DefinePDF (LHA)*)


(* ::Subsection:: *)
(*The pdf function accepts values of x, q, parton #, and pdf data set in the following syntax: pdf[x,q,ipart,{dataset}].*)



(*plist: produced the value of the PDF at x, Q ,for flavor, and pdfDTabledata is the list*)
plist[x_?NumericQ,q_?NumericQ,flav_?IntegerQ,list_]:= fullinterp[x,q,flav,list];


(*pdfX: the internal version of the pdfFunctionLHA, input is set, flavor, x, and Q, output is the pdf*)
pdfX[iset_?IntegerQ,ipart_?IntegerQ,x_?NumericQ,q_?NumericQ]:=
	Module[{list,tpdf,nfmx,nfval,miscList,nfl},
	list=pdfTableData[iset];
	If[list==False,Return[Null]];  
	miscList=list[[4]];
	{nfl,nfmx,nfval}=miscList;

     ipartmod=flavorPatch[ipart,nfl]; (* THIS IS A PATCH: NEED TO MAKE MORE FLEXIBLE *)

	Return[plist[x,q,ipartmod,list]];
  ];

(* OLD VERSION 
(*flavorshift: shifts the flavor numbers the user would input to refer to the correct flavor data in the grid*)
flist11={1,2,3,4,5,11,6,7,8,9,10};(*The numbers refer to the column number in the .dat file and the order refers to the placement in numerical convention order -5\[Rule]5 with the gluon at 0*)
flist13={1,2,3,4,5,6,13,7,8,9,10,11,12};(*The numbers refer to the column number in the .dat file and the order refers to the placement in numerical convention order -5\[Rule]5 with the gluon at 0*)
*)

(* flavorshift[nf_]:=flist[[nf+6]];  OLD VERSION *)

(*  PATCH FOR FLAVORS:  *)
flavorPatch[ipart_,nfl_]:=Module[{ipartmod,nshift,flavorshift,flist,flist11={1,2,3,4,5,11,6,7,8,9,10},flist13={1,2,3,4,5,6,13,7,8,9,10,11,12}},
    If[nfl==13,flist=flist13,flist=flist11];
    nshift=(nfl+1)/2;
    flavorshift[nf_]:=flist[[nf+nshift]];
	ipartmod=flavorshift[ipart];
     Return[ipartmod];
]




(*pdfFunctionLHA: user version of the pdfX function deffinition*)
pdfFunctionLHA[iset_,ipart_,x_,q_] := pdfX[iset,ipart,x,q];



(* ::Section:: *)
(*Alpha_s implimentation*)



(*getalphadata: constructs the data for the alphaS interoplation*)
getalphadata[qi_, alphalist_] := 
 Transpose[{alphalist[[1, qi - 2 ;; qi + 1]], 
   alphalist[[2, qi - 2 ;; qi + 1]]}
];


(*alphaS: input is set number and value of q, ouput is interpolated value of alphpaS at that point*)
alphaS[q_] := Module[{qini, alfa, data, output},
   alfa = alphalist;
   qini = findQindex[q, alfa[[1]]];
   data = getalphadata[qini, alfa];
   output = interpol4[q, data];
   Return[output]
   ];


(*pdfAlphaSLHA: user available version of alphaS function. Will return various prited messages if it cannot interpolate*)
pdfAlphaSLHA[q_] := Module[{},
    output = alphaS[q];
    Return[output];
   ];






(*pdfFlavorTMP: a list (internal) of names of partons in LHA scheme*)
pdfFlavorTMP={
	"tbar",
	"bbar",
	"cbar",
	"sbar",
	"ubar",
	"dbar",
	"gluon",
	"down",
	"up",
	"strange",
	"charm",
	"bottom",
	"top"
	};


(*pdfFlavorLHA:  User available function-->input is parton number in LHA scheme, output is string that is name of parton*)
pdfFlavorLHA[i_]:=If[i==21,Print["For LHA, gluon is denoted 21, however for this package the gluon is denoted as 0"],
						If[i!=21,pdfFlavorTMP[[i+7]],pdfFlavorTMP[[7]]
						]
					];


(* ::Section:: *)
(*pdfGet Functions:*)


(*pdfGetTableLHA: input is set number, output is grid values from the associated PDF*)
pdfGetTableLHA[iset_]:=pdfTableData[iset][[3]];

(*pdfGetValueLHA: input is set number, flavor indicie, q indicie, x indicie, output is grid values from the associated PDF*)
pdfGetValueLHA[iset_,flavor_,q_,x_]:=pdfTableData[iset][[3,flavorshift[flavor],q,x]];

(*pdfGetXlistLHA: input is set number, output is x grid from the associated PDF*)
pdfGetXlistLHA[iset_]:=pdfTableData[iset][[1]];


(*pdfGetQlistLHA: input is set number, output is Q grid from the associated PDF*)
pdfGetQlistLHA[iset_]:=pdfTableData[iset][[2]];


(*pdfXminLHA: input is set number, ouput is the minimum value in the x grid in the assosiated PDF*)
pdfXminLHA[iset_] := Module[{value},
value=pdfxmin[iset];
If[value==False,Return[Null]];
Return[value];
];

(*pdfGetInfoLHA: no input, prints the contents of the info file*)
pdfGetInfoLHA[]:= Module[{},

For[j=1,j<=Length[readlist],j++,
Print[readlist[[j]],": ", datalist[[j]]];
];

Print["Flavors: ",flavorlist];
Print["AlphaQlist: ",AlphaQlist];
Print["AlphaValslist: ",AlphaValslist];
];

(*pdfGetQlistLHA: no input, output is the list of AlphaS values*)
pdfGetAlphaValuesLHA[]:=AlphaValslist;



(* ::Section:: *)
(*End Package*)


End[];  (* End Private Context *)

EndPackage[]; (* End Package Context *)
